% =========================================================================
% Title       : Wrapper for Iterative MIMO Detectors
% File        : detector.m
% -------------------------------------------------------------------------
% Description :
%   This file performs preprocessing and calls the chosen iterative MIMO
%   detector. This version of the simulator only includes a reduced-
%   complexity SISO MMSE-PIC algorithm and the SISO STS-SD but can be 
%   extended easily with other detectors. 
% ------------------------------------------------------------------------- 
% Revisions   :
%   Date       Version  Author  Description
%   14-sep-07  2.0      studer  adapted for student environment
%   05-feb-07  1.2      studer  some detectors require intrinsic feedback
%   22-sep-06  1.1      studer  LLR clipping added
%   05-may-06  1.0      studer  partially adapted from ofdm_sim
% -------------------------------------------------------------------------
%   (C) 2006-2010 Communication Theory Group                      
%   ETH Zurich, 8092 Zurich, Switzerland                               
%   Author: Dr. Christoph Studer (e-mail: studerc@nari.ee.ethz.ch)     
% =========================================================================

function [LLR_E1,complexity_log] = detector(TxRx,yN,HN,SNR,LLR_A1,LLR_P1)
  
  % -- initialization 
  complexity_log = 0; 
  LLR_E1_shaped = zeros(TxRx.Ntx,TxRx.Nused*TxRx.Modulation_order); % extrinsic a-posteriori information
  LLR_A1_shaped = reshape(LLR_A1,TxRx.Nused*TxRx.Modulation_order,TxRx.Ntx).'; 
  LLR_P1_shaped = reshape(LLR_P1,TxRx.Nused*TxRx.Modulation_order,TxRx.Ntx).';
  
  % -- perform detection over the N used OFDM tones
  for n=1:TxRx.Nused
     
    % -- extract data for n-th tone
    H = HN(:,:,TxRx.ToneMap(n));
    y = yN(:,n);    
    APP_LLR = LLR_A1_shaped(:,(n-1)*TxRx.Modulation_order+1:n*TxRx.Modulation_order); % extrinsic a-priori input
    PP_LLR  = LLR_P1_shaped(:,(n-1)*TxRx.Modulation_order+1:n*TxRx.Modulation_order); % intrinsic a-priori input
    
    % -- preprocessing
    switch (TxRx.Detector.Preprocessing)
      case 'off'
        % don't do anything (e.g., for SISO MMSE-PIC)
      otherwise,
        % -- perform preprocessing (e.g., for SISO STS-SD)
        [Q,R,order] = preprocessing(TxRx,H,SNR);
        y_hat = Q'*y;        
    end
    
    % -- choose the specified soft-in soft-out detector
    switch (TxRx.Detector.Type) 
      case 'SISO_MMSE_PIC',  % reduced-complexity SISO MMSE-PIC 
        [LLR_E1_data,complexity] = MMSE_PIC(TxRx,H,y,SNR,PP_LLR);
      case 'SISO_STS_SD', % SISO STS-SD detector [Studer and Boelcskei, ISIT 2008]
        [LLR_E1_data,complexity] = SISO_STS_SD(TxRx,Q,R,y_hat,order,SNR,APP_LLR);     
      otherwise,
        error('No valid TxRx.Detector.Type in detector.m')
    end;    
    
    % -- re-order LLR data
    LLR_E1_shaped(:,(n-1)*TxRx.Modulation_order+1:n*TxRx.Modulation_order) = LLR_E1_data;
    complexity_log  = complexity_log + complexity; % assess complexity       
    
  end

  % -- reshape a-posteriori LLRs
  LLR_E1 = reshape(LLR_E1_shaped.',1,TxRx.Ntx*TxRx.Nused*TxRx.Modulation_order);      
  
return
