% =========================================================================
% Title       : Reduced-Complexity SISO MMSE PIC
% File        : MMSE_PIC.m
% -------------------------------------------------------------------------
% Description :
%   Reduced-complexity soft-input soft-output MMSE parallel interference 
%   cancellation (PIC) MIMO detector for iterative MIMO decoding.
% ------------------------------------------------------------------------- 
% Revisions   :
%   Date       Version  Author  Description
%   14-sep-07   1.2     studer  full a priori input required
%   16-jul-06   1.1     studer  reduced complexity variant
%   29-jun-06   1.0     studer  file created
% -------------------------------------------------------------------------
%   (C) 2006-2010 Communication Theory Group                      
%   ETH Zurich, 8092 Zurich, Switzerland                               
%   Author: Dr. Christoph Studer (e-mail: studerc@nari.ee.ethz.ch)     
% =========================================================================

function [LLR_E_data,complexity] = MMSE_PIC(TxRx,H,y,SNR,PP_LLR)
     
  complexity = -1;

  % -- initialization
  LLR_D_data = zeros(TxRx.Ntx,TxRx.Modulation_order);
  LLR_E_data = zeros(TxRx.Ntx,TxRx.Modulation_order);
  bin_array = sign(mexde2bi([0:2^(TxRx.Modulation_order)-1],TxRx.Modulation_order)-0.5);   
  N0 = 1/SNR; 
  
  % -- calculate soft estimates
  for n=[1:TxRx.Ntx]
    Distances = 0.5*bin_array*PP_LLR(n,:)'; Distances = Distances+min(Distances); % posterior app
    Probs = exp(Distances); Probs = Probs'/sum(Probs);       
    sEst(n,1) = sum(Probs.*TxRx.Constellations_norm);
    eEst(n,1) = sum(Probs.*abs(sEst(n)*ones(1,length(TxRx.Constellations_norm))-TxRx.Constellations_norm).^2);      
  end
    
  % -- one-time inverse (numerically stable version)
  G = inv(H'*H*diag(eEst)+eye(TxRx.Ntx)*N0)*H';  
  
  % -- MMSE detection main loop
  for n=[TxRx.Ntx:-1:1]

    % -- interference cancellation
    Hslice = H; Hslice(:,n) = []; % remove n-th column
    sslice = sEst; sslice(n,:) = []; % remove n-th dimension
    eslice = eEst; eslice(n,:) = []; % ^^
    y_tilde = y - Hslice*sslice; % cancel known interference
        
    % -- compute noise-plus-interference terms
    NPI = real(G(n,:)*(Hslice*diag(eslice)*Hslice'+eye(TxRx.Ntx)*N0)*G(n,:)');    
    Metrics = abs(G(n,:)*y_tilde-G(n,:)*H(:,n)*TxRx.Constellations_norm).^2/NPI;
        
    % -- compute max-log LLRs
    for b=[1:TxRx.Modulation_order]
      pMin = inf; nMin = inf;
      for z=[1:2^(TxRx.Modulation_order)]          
        if (bin_array(z,b)==1)
          pMin = min(pMin,Metrics(z));
        else
          nMin = min(nMin,Metrics(z));          
        end                                
      end            
      LLR_D_data(n,b) = nMin-pMin;
    end 
             
  end

  % -- output intrinsic information
  LLR_E_data = LLR_D_data; 
return