% =========================================================================
% Title       : Error Simulation Environment for Iterative MIMO Decoding
% File        : err_sim.m
% -------------------------------------------------------------------------
% Description :
%   Main simulation file for iterative MIMO decoding according to Hochwald 
%   and ten Brink, 2003. This file handles initialization, performs 
%   Monte-Carlo simulation loops, and stores the results.
% ------------------------------------------------------------------------- 
% Revisions   :
%   Date       Version  Author  Description
%   01-dec-08  1.2      studer  major cleanup and speedup
%   14-sep-06  1.1      studer  enhanced with iterative MIMO decoding
%   10-aug-06  1.0      studer  adapted from ofdm_sim environment
% -------------------------------------------------------------------------
%   (C) 2006-2010 Communication Theory Group                      
%   ETH Zurich, 8092 Zurich, Switzerland                               
%   Author: Dr. Christoph Studer (e-mail: studerc@nari.ee.ethz.ch)     
% =========================================================================

function err_sim(TxRx,RunID)
  
  % -- initialization 
  [TxRx,Result] = init(TxRx,RunID);
  K_list = 1:length(TxRx.Sim.SNR_list);
    
  % -- main simulation loop (packets)
  tic;
  for pack=1:TxRx.Sim.nr_of_packets

    % -- generate channel matrices according to specified channel model
    HN = channel(TxRx); 

    % -- transmit OFDM symbols
    PERROR = zeros(length(TxRx.Sim.SNR_list),TxRx.Iterations);
    for symb=1:TxRx.Sim.nr_of_symbols
    
      % -- transmitter: generates bits, maps to symbol vectors, 
      % -- and transmits symbol vectors over (noiseless) channel 
      [data_bin,xN] = tx(TxRx,HN);
      
      % -- generate noise realization
      noise = sqrt(0.5)*(randn(TxRx.Nrx,TxRx.Nused)+1i*randn(TxRx.Nrx,TxRx.Nused)); 
    
      for k=K_list
        SNR = TxRx.Sim.SNR_list(k);
           
        % -- add Gaussian noise
        yN = xN + noise*sqrt(1/SNR);

        % -- initialize a-priori and a-posteriori information
        % some detectors need intrinsic priors (e.g., SISO MMSE-PIC)
        LLR_A1 = zeros(1,TxRx.Ntx*TxRx.Nused*TxRx.Modulation_order); % extrinsic a-priori input for detector
        LLR_P1 = zeros(1,TxRx.Ntx*TxRx.Nused*TxRx.Modulation_order); % intrinsic prior information for detector
         
        % -- BEGIN iterative MIMO LOOP
        for iteration=1:TxRx.Iterations     
          
          % -- wrapper for soft-in soft-out MIMO detectors
          [LLR_E1,complexity] = detector(TxRx,yN,HN,SNR,LLR_A1,LLR_P1);  
     
          % -- wrapper for channel decoders
          [LLR_A1,LLR_P1,data_bin_hat] = decoder(TxRx,LLR_E1);           
        
          % -- compute coded BER/FER
          tmp = sum(abs(data_bin_hat-data_bin));
          %[ complexity , tmp ] 
          Result.BER(k,iteration) = Result.BER(k,iteration) + tmp;
          Result.SER(k,iteration) = Result.SER(k,iteration) + (tmp>0);   
          if (tmp>0)
            PERROR(k,iteration) = 1;
          end
          Result.complexity(k,iteration) = Result.complexity(k,iteration) + complexity;
          
        end  
              
      end % end SNR loop

    end % end symbol loop
 
    % -- count packet errors
    Result.PER = Result.PER + PERROR;
    Result.Npackets(K_list) = Result.Npackets(K_list) + 1;
    K_list(find(Result.PER(K_list,TxRx.Iterations)>TxRx.Sim.MaxPER)) = [];
    
    % -- save intermediate results
    Result.total_time = Result.total_time + toc;
    save_results(TxRx,RunID,Result,pack);
    
    % -- show # of elapsed packets
    disp(sprintf('\nPacket %i of %i transmitted',pack,TxRx.Sim.nr_of_packets))
      
  end % end packet loop

  % -- simulation completed!
  disp(' =====================================================');
  disp(sprintf(' ==================== Total Simulation Time: %f sec. ',Result.total_time));
  disp(' =====================================================');

  % -- store FINAL results to disk
  save_results(TxRx,RunID,Result,TxRx.Sim.nr_of_packets);
  
return

% -------------------------------------------------------------------------
% -- save (intermediate/final) results to disk
% -------------------------------------------------------------------------
function save_results(TxRx,RunID,Res,NrOfSimPackets) 
  % -- compose simulation results structure
  Results.TxRx = TxRx;
  Results.Npackets = Res.Npackets;  
  Results.RunName = TxRx.Sim.name;
  Results.RunID = RunID;
  Results.simulationTime = Res.total_time; 
  Results.NrOfSimPackets = NrOfSimPackets; % store actual number of simulated channels
  if (NrOfSimPackets==TxRx.Sim.nr_of_packets)
    Results.Final = 'yes';
  else
    Results.Final = 'no';
  end
  % -- for each iteration
  for iteration=1:TxRx.Iterations    
    Results.BER = Res.BER(:,iteration)./(Res.Npackets*TxRx.Sim.nr_of_symbols*TxRx.Nused*TxRx.N*TxRx.Modulation_order*TxRx.Code.Rate);  
    Results.SER = Res.SER(:,iteration)./(Res.Npackets*TxRx.Sim.nr_of_symbols);
    Results.PER = Res.PER(:,iteration)./Res.Npackets;
    % -- compute cumulative complexity
    Results.complexity = sum(Res.complexity(:,1:iteration),2)./(Res.Npackets*TxRx.Sim.nr_of_symbols);
    % -- store Results structure to a separate file
    Results.FileName = sprintf('results/%s_I%d_%d.mat',TxRx.Sim.name,iteration,RunID);
    save(Results.FileName,'Results');
  end
  
return