% =========================================================================
% Title       : Initialization of Simulation 
% File        : init.m
% -------------------------------------------------------------------------
% Description :
%   This file initializes random seeds, prepares constellation symbols 
%   with the (Gray) mapping, generates the interleaver parameters, prepares
%   puncturing patterns, and SNR values. Note: this function requires the
%   communication toolbox for the poly2trellis command.
% -------------------------------------------------------------------------
% Revisions   :
%   Date       Version  Author  Description
%   17-jun-09  1.0      studer  file adapted from large sim-environment
% -------------------------------------------------------------------------
%   (C) 2006-2010 Communication Theory Group                      
%   ETH Zurich, 8092 Zurich, Switzerland                               
%   Author: Dr. Christoph Studer (e-mail: studerc@nari.ee.ethz.ch)     
% =========================================================================              

function [TxRx,Result] = init(TxRx,RunID)

  % -- initialization of random seeds
  randn('seed',RunID);
  rand('seed',RunID); 
   
  % -- construct IEEE 802.11a-compliant mapping/constellation points
  switch (TxRx.Modulation_order)
    case 1,
      TxRx.Constellations = [ -1 1 ];
    case 2,
      TxRx.Constellations = [ -1-1j,-1+1j, ...
                              +1-1j,+1+1j ];
    case 4,
      TxRx.Constellations = [ -3-3j,-3-1j,-3+3j,-3+1j, ...
                              -1-3j,-1-1j,-1+3j,-1+1j, ...
                              +3-3j,+3-1j,+3+3j,+3+1j, ...
                              +1-3j,+1-1j,+1+3j,+1+1j ];
    case 6,
      TxRx.Constellations = [ -7-7j,-7-5j,-7-1j,-7-3j,-7+7j,-7+5j,-7+1j,-7+3j, ...
                              -5-7j,-5-5j,-5-1j,-5-3j,-5+7j,-5+5j,-5+1j,-5+3j, ...
                              -1-7j,-1-5j,-1-1j,-1-3j,-1+7j,-1+5j,-1+1j,-1+3j, ...
                              -3-7j,-3-5j,-3-1j,-3-3j,-3+7j,-3+5j,-3+1j,-3+3j, ...
                              +7-7j,+7-5j,+7-1j,+7-3j,+7+7j,+7+5j,+7+1j,+7+3j, ...
                              +5-7j,+5-5j,+5-1j,+5-3j,+5+7j,+5+5j,+5+1j,+5+3j, ...
                              +1-7j,+1-5j,+1-1j,+1-3j,+1+7j,+1+5j,+1+1j,+1+3j, ...
                              +3-7j,+3-5j,+3-1j,+3-3j,+3+7j,+3+5j,+3+1j,+3+3j ];
    otherwise,  
      error('Modulation order for 11a mapping not supported.')
  end
  % -- normalize power such that Ex[|s|^2]=1/TxRX.Ntx
  const_power = sum(abs(TxRx.Constellations).^2)/length(TxRx.Constellations);
  TxRx.Constellations_norm = TxRx.Constellations/sqrt(TxRx.Ntx*const_power);
  
  % -- count number of used OFDM tones
  TxRx.Nused = length(TxRx.ToneMap);
  
  % -- generate code and interleaver data
  TxRx.Code.trellis = poly2trellis(TxRx.Code.K, TxRx.Code.generators); % requires communication toolbox
  TxRx.Code.InterleaverPerm = randperm(TxRx.Nused*TxRx.Modulation_order*TxRx.Ntx); % random interleaver

  % -- puncturing according to IEEE 802.11n
  switch (TxRx.Code.Rate) % total code rate (after puncturing)
    case 1/2, % -- no puncturing
      TxRx.Code.Puncturing.Period = 1;
      TxRx.Code.Puncturing.Pattern = [ 1 ];
    case 3/4,
      TxRx.Code.Puncturing.Period = 18;
      TxRx.Code.Puncturing.Pattern = [ 1 2 3 6 7 8 9 12 13 14 15 18 ];
    case 2/3, 
      TxRx.Code.Puncturing.Period = 12;
      TxRx.Code.Puncturing.Pattern = [ 1 2 3 5 6 7 9 10 11];
    case 5/6,
      TxRx.Code.Puncturing.Period = 10;
      TxRx.Code.Puncturing.Pattern = [ 1 2 3 6 7 10 ];
    otherwise,  
      error('Rate not supported')    
  end
  codedBits = length(TxRx.ToneMap)*TxRx.Ntx*TxRx.Modulation_order;     
  patLen = length(TxRx.Code.Puncturing.Pattern); 
  for i=1:ceil(codedBits/patLen)       
    TxRx.Code.Puncturing.Index(patLen*(i-1)+1:patLen*i) = ...
          (i-1)*TxRx.Code.Puncturing.Period+TxRx.Code.Puncturing.Pattern;   
  end
  TxRx.Code.Puncturing.Index(:,codedBits+1:end) = [];
  TxRx.Code.PADbits = ceil(TxRx.Ntx*TxRx.Nused*TxRx.Modulation_order*TxRx.Code.Rate)-floor(TxRx.Ntx*TxRx.Nused*TxRx.Modulation_order*TxRx.Code.Rate);
  
  % -- Compute Eb/N0 and SNR in dB
  TxRx.Sim.EBN0dB_list = TxRx.Sim.SNR_dB_list + 10*log10(TxRx.Nrx/TxRx.Ntx/TxRx.Modulation_order/TxRx.Code.Rate);
  TxRx.Sim.SNR_list = 10.^(TxRx.Sim.SNR_dB_list/10);
  
  % -- prepare Result structure
  Result.total_time = 0;
  Result.BER = zeros(length(TxRx.Sim.SNR_list),TxRx.Iterations); % bit error rate
  Result.SER = zeros(length(TxRx.Sim.SNR_list),TxRx.Iterations); % OFDM-symbol error rate
  Result.PER = zeros(length(TxRx.Sim.SNR_list),TxRx.Iterations); % packet error rate
  Result.complexity = zeros(length(TxRx.Sim.SNR_list),TxRx.Iterations);
  Result.Npackets = zeros(length(TxRx.Sim.SNR_list),1);
  
return
  